#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2010, 2011 Richard Dymond (rjdymond@gmail.com)
#
# This file is part of SkoolKit.
#
# SkoolKit is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# SkoolKit is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# SkoolKit. If not, see <http://www.gnu.org/licenses/>.

import sys

def get_word(word):
    return chr(word % 256) + chr(word // 256)

def get_bytes(*bytes):
    return ''.join([chr(b) for b in bytes])

def get_parity(block):
    parity = 0
    for c in block[2:]:
        parity ^= ord(c)
    return chr(parity)

def get_basic_loader(title):
    header = get_word(19)          # Length of header block
    header += chr(0)               # Header block marker
    header += chr(0)               # BASIC program follows
    header += title[:10].ljust(10) # Title padded with spaces up to 10 chars
    header += get_word(27)         # Length of entire data block
    header += get_word(10)         # LINE number to RUN after LOADing
    header += get_word(27)         # Length of BASIC program only
    header += get_parity(header)

    data = get_word(29)            # Length of data block
    data += chr(255)               # Data block marker
    data += chr(0) + chr(10)       # Line 10
    data += get_word(5)            # Length of line 10
    data += chr(239)               # LOAD
    data += chr(34) + chr(34)      # ""
    data += chr(175)               # CODE
    data += chr(13)                # ENTER
    data += chr(0) + chr(20)       # Line 20
    data += get_word(14)           # Length of line 20
    data += chr(249)               # RANDOMIZE
    data += chr(192)               # USR
    data += "23296"                # 23296
    data += chr(14)                # Floating-point number marker
    data += get_bytes(0, 0)        # 23296 in
    data += get_word(23296)        # floating-point
    data += chr(0)                 # form
    data += chr(13)                # ENTER
    data += get_parity(data)

    return header + data

def get_data_loader(title, org, length, start):
    header = get_word(19)          # Length of header block
    header += chr(0)               # Header block marker
    header += chr(3)               # CODE block follows
    header += title[:10].ljust(10) # Title padded with spaces up to 10 chars
    header += get_word(18)         # Length of data block
    header += get_word(23296)      # Start address
    header += get_word(0)          # Unused
    header += get_parity(header)

    data = get_word(20)                        # Length of data block
    data += chr(255)                           # Data block marker
    data += get_bytes(221, 33) + get_word(org) # LD IX,ORG
    data += chr(17) + get_word(length)         # LD DE,LENGTH
    data += get_bytes(62, 255)                 # LD A,255
    data += get_bytes(55)                      # SCF
    data += get_bytes(221, 249)                # LD SP,IX
    data += get_bytes(205, 86, 5)              # CALL 1366
    data += chr(195) + get_word(start)         # JP START
    data += get_parity(data)

    return header + data

def parse_options(options, binfile, binlength):
    org = start = 65536 - binlength
    suffix = '.bin'
    if binfile.endswith(suffix):
        prefix = binfile[:-len(suffix)]
    else:
        prefix = binfile
    tapfile = prefix + ".tap"
    i = 0
    while i < len(options):
        option = options[i]
        value = options[i + 1]
        if option == '-o':
            org = int(value)
        elif option == '-s':
            start = int(value)
        elif option == '-t':
            tapfile = value
        i += 2
    return org, start, tapfile

#
# Begin
#
if len(sys.argv) < 2:
    sys.stderr.write("Usage: %s [options] file.bin\n\n" % sys.argv[0])
    sys.stderr.write("Options:\n")
    sys.stderr.write("  -o ORG      Set the origin (default: 65536 - length of file.bin)\n")
    sys.stderr.write("  -s START    Set the start address to JP to (default: ORG)\n")
    sys.stderr.write("  -t TAPFILE  Set the TAP filename (default: file.tap)\n")
    sys.exit(1)

binfile = sys.argv[-1]
f = open(binfile, 'rb')
ram = f.read()
f.close()

length = len(ram)
org, start, tapfile = parse_options(sys.argv[1:-1], binfile, length)
data = get_word(length + 2) # Length of data block
data += chr(255)            # Data block marker
data += ram                 # Data
data += get_parity(data)

tap_data = get_basic_loader(binfile)
tap_data += get_data_loader(binfile, org, length, start)
tap_data += data

f = open(tapfile, 'w')
f.write(tap_data)
f.close()
