# -*- coding: utf-8 -*-

# Copyright 2008-2011 Richard Dymond (rjdymond@gmail.com)
#
# This file is part of SkoolKit.
#
# SkoolKit is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# SkoolKit is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# SkoolKit. If not, see <http://www.gnu.org/licenses/>.

import sys
import textwrap
import re
import time
from skoolparser import TableParser

def parse_int(num_str):
    if num_str[0] in '$#&':
        return int(num_str[1:], 16)
    return int(num_str)

def get_num_str(num_str):
    return num_str[1:] if num_str.startswith('$') else num_str

class SkoolWriter:
    def __init__(self, parser, crlf, tab, lower, instr_width):
        self.parser = parser

        # Build a label dictionary
        self.labels = {}
        for address, instructions in parser.instructions.items():
            label = instructions[0].asm_label
            if label:
                self.labels[address] = label
        if self.labels:
            self.base_address = min([address for address in self.labels.keys()])
        else:
            self.base_address = int(self.parser.memory_map[0].instructions[0].address)

        self.lower = lower

        # Line terminator and indent
        self.end = '\r\n' if crlf else '\n'
        self.indent = '\t' if tab else '  '

        # Field widths (line = indent + instruction + ' ; ' + comment)
        indent_width = 8 if tab else len(self.indent)
        self.instr_width = instr_width
        self.line_width = 79 # Instead of 80 (to be diff-friendly)

        # Comment widths depend on the other field widths
        self.comment_width = self.line_width - 3 - self.instr_width - indent_width
        self.desc_width = self.line_width - 2
        self.table_writer = TableWriter(self.desc_width, 10)

        # Markers
        self.table_marker = TableParser.TABLE_MARKER
        self.table_end_marker = TableParser.TABLE_END_MARKER
        self.udg_table_marker = '#UDGTABLE'
        self.p_marker = '#P'
        self._create_macros()

    def _warn(self, s):
        sys.stderr.write('WARNING: %s\n' % s)

    def _error(self, s):
        sys.stderr.write('ERROR: %s\n' % s)
        sys.exit(1)

    def write(self):
        self.print_header(self.parser.header)
        for entry in self.parser.memory_map:
            first_instruction = entry.instructions[0]
            org = first_instruction.org
            if org:
                self.write_line('%s%s %s' % (self.indent, 'org' if self.lower else 'ORG', org))
                self.write_line('')
            self.entry = entry
            self.print_entry()
            self.write_line('')

    def print_header(self, header):
        if header:
            for line in header:
                self.write_line(('; %s' % self.strip_macros(line)).rstrip())
            self.write_line('')

    def print_entry(self):
        self.print_comment_lines(self.entry.description)
        if self.entry.details:
            self.write_line(';')
            self.print_comment_lines(self.entry.details)
        if self.entry.is_routine() and self.entry.registers:
            self.write_line(';')
            for reg in self.entry.registers:
                self.write_line('; %s %s' % (reg.name, self.strip_macros(reg.contents)))
        self.print_instructions()
        if self.entry.end_comment:
            self.print_comment_lines(self.entry.end_comment)

    def write_line(self, s):
        sys.stdout.write('%s%s' % (s, self.end))

    def replace_address(self, params, anchor, link_text):
        if link_text:
            return link_text
        addr_str = params[0][:5]
        if params[0][5:].startswith('@'):
            return addr_str
        address = parse_int(addr_str)
        label = self.labels.get(address)
        if label is None:
            if address >= self.base_address:
                self._warn('Could not convert address %s to label' % addr_str)
            instructions = self.parser.instructions.get(address)
            if instructions:
                label = instructions[0].addr_str
            else:
                label = addr_str[1:] if addr_str.startswith('$') else addr_str
        return label.lower() if self.lower else label

    def parse_macro(self, text, index, chars=''):
        end = index
        if text[index] == '$':
            end += 1
        valid_chars = '#' + chars
        while end < len(text) and (text[end].isalnum() or text[end] in valid_chars):
            end += 1
        param_str = text[index:end].partition('#')
        params = param_str[0].split(',')
        anchor = param_str[2]
        link_text = ''
        if end < len(text) and text[end] == '(':
            closing = text.index(')', end)
            link_text = text[end + 1:closing]
            end = closing + 1
        return end, params, anchor, link_text

    def _create_macros(self):
        # 'parse_int(params[0])' is used where an integer is expected so that
        # we error out if one is not found
        self.macros = {}
        self.macros['#BUG'] = lambda param, anchor, link_text: link_text or 'bug'
        self.macros['#C'] = lambda param, anchor, link_text: '(c)'
        self.macros['#CALL'] = None
        self.macros['#D'] = lambda params, anchor, link_text: self.parser.entries[parse_int(params[0])].description
        self.macros['#EREFS'] = None
        self.macros['#FACT'] = None
        self.macros['#FONT'] = None
        self.macros['#LINK'] = lambda param, page_id, link_text: link_text
        self.macros['#NAME'] = lambda params, anchor, link_text: ''
        self.macros['#P'] = lambda params, anchor, link_text: ''
        self.macros['#POKE'] = None
        self.macros['#POKES'] = None
        self.macros['#POPS'] = None
        self.macros['#PUSHS'] = None
        self.macros['#R'] = self.replace_address
        self.macros['#REFS'] = None
        self.macros['#REG'] = lambda params, anchor, link_text: params[0] if self.lower else params[0].upper()
        self.macros['#SCR'] = None
        self.macros['#SPACE'] = lambda params, anchor, link_text: ''.ljust(parse_int(params[0]))
        self.macros['#TABLE'] = None
        self.macros['#UDG'] = None
        self.macros['#UDGARRAY'] = None
        self.macros['#UDGTABLE'] = None

    def strip_macros(self, text):
        if text.find('#') < 0:
            return text

        # Remove #TABLE and #UDGTABLE macros
        text = text.replace('#TABLE', '$TABLE')
        text = text.replace('#UDGTABLE', '$UDGTABLE')

        while True:
            search = re.search('#[A-Z]+', text)
            if not search:
                break
            marker = search.group()
            if not marker in self.macros:
                self._error('Found unknown macro: %s' % marker)
            repf = self.macros.get(marker)
            if repf:
                start, index = search.span()
                valid_chars = ''
                if marker == '#REG':
                    valid_chars = "'"
                elif marker == '#R':
                    valid_chars = '@'
                elif marker == '#LINK':
                    valid_chars = ':'
                end, params, anchor, link_text = self.parse_macro(text, index, valid_chars)
                rep = repf(params, anchor, link_text)
                text = text.replace(text[start:end], str(rep), 1)
            else:
                self._error('Found unsupported macro: %s' % marker)

        # Reinstate #TABLE and #UDGTABLE macros
        text = text.replace('$TABLE', '#TABLE')
        text = text.replace('$UDGTABLE', '#UDGTABLE')

        return text.strip()

    def find_table_markers(self, table_indexes, text, marker):
        index = 0
        while text.find(marker, index) >= 0:
            table_index = text.index(marker, index)
            table_end_index = text.index(self.table_end_marker, table_index) + len(self.table_end_marker)
            table_indexes.append(table_index)
            table_indexes.append(table_end_index)
            index = table_end_index

    def extract_blocks(self, text):
        # Find paragraph markers
        p_indexes = []
        index = 0
        while text.find(self.p_marker, index) >= 0:
            p_index = text.index(self.p_marker, index)
            p_indexes.append(p_index)
            index = p_index + len(self.p_marker)

        # Find table markers
        table_indexes = []
        self.find_table_markers(table_indexes, text, self.table_marker)
        self.find_table_markers(table_indexes, text, self.udg_table_marker)

        # Extract blocks
        blocks = []
        all_indexes = [0]
        all_indexes += p_indexes + table_indexes
        all_indexes.sort()
        all_indexes.append(len(text))
        for i in range(len(all_indexes) - 1):
            start = all_indexes[i]
            if start in p_indexes:
                start += len(self.p_marker)
            end = all_indexes[i + 1]
            block = text[start:end].strip()
            if block and not block.startswith(self.udg_table_marker):
                blocks.append(block)

        return blocks

    def format(self, text, width):
        lines = []
        for block in self.extract_blocks(text):
            if block.startswith(self.table_marker):
                table_lines = self.table_writer.format_table(self.strip_macros(block))
                table_width = max([len(line) for line in table_lines])
                if table_width > width:
                    self._warn('Table in entry at %s is %i characters wide' % (self.entry.address, table_width))
                lines.extend(table_lines)
            else:
                lines.extend(textwrap.wrap(self.strip_macros(block), width))
            if lines[-1]:
                lines.append('')
        return lines[:-1]

    def print_comment_lines(self, comment, instruction=None):
        for line in self.format(comment, self.desc_width):
            uaddress = self.find_unconverted_address(line)
            if uaddress and not (line.startswith('| ') and line.endswith(' |')):
                # Print a warning only if the unconverted address appears
                # outside a table row (there are many unconvertible addresses in
                # table rows that would otherwise flood the warning output, so
                # we ignore them)
                if instruction:
                    if not instruction.ignoremrcua:
                        self._warn('Comment above %s contains address (%i) not converted to a label:\n; %s' % (instruction.address, uaddress, line))
                else:
                    self._warn('Comment contains address (%i) not converted to a label:\n; %s' % (uaddress, line))
            self.write_line(('; %s' % line).rstrip())

    def print_instruction_prefix(self, instruction):
        mid_routine_comment = instruction.get_mid_routine_comment()
        if mid_routine_comment:
            self.print_comment_lines(mid_routine_comment, instruction)
        asm_label = instruction.asm_label
        if asm_label:
            if self.lower:
                asm_label = asm_label.lower()
            self.write_line("%s:" % asm_label)

    def find_unconverted_address(self, text):
        search = re.search('[1-9][0-9][0-9][0-9][0-9]', text)
        if search:
            start, end = search.span()
            if (start == 0 or text[start - 1] == ' ') and (end == len(text) or not text[end].isalnum()):
                uaddress = int(search.group())
                if uaddress >= self.base_address:
                    return uaddress

    def print_instructions(self):
        i = 0
        rowspan = 0
        lines = []
        new_instruction = True
        instructions = self.entry.instructions
        while i < len(instructions) or lines:
            instruction = instructions[i] if i < len(instructions) else None

            # Deal with remaining comment lines or rowspan on the previous
            # instruction
            if lines or rowspan > 0:
                if rowspan > 0:
                    self.print_instruction_prefix(instruction)
                    operation = instruction.operation
                    rowspan -= 1
                    i += 1
                else:
                    operation = ''

                if lines:
                    line_comment = lines.pop(0)
                    oline = '%s%s ; %s' % (self.indent, operation.ljust(self.instr_width), line_comment)
                else:
                    line_comment = ''
                    oline = '%s%s' % (self.indent, operation)
                self.write_line(oline)
                if not ignoreua:
                    uaddress = self.find_unconverted_address(line_comment)
                    if uaddress:
                        self._warn('Comment at %s contains address (%i) not converted to a label:\n%s' % (iaddress, uaddress, oline))
                if lines and len(operation) > self.instr_width:
                    self._warn('Operation is %i characters long:\n%s' % (len(operation), oline))
                elif len(oline) > self.line_width:
                    self._warn('Line is %i characters long:\n%s' % (len(oline), oline))
                continue

            ignoreua = instruction.ignoreua
            iaddress = instruction.address

            if instruction.comment:
                rowspan = instruction.comment.rowspan
                lines = textwrap.wrap(self.strip_macros(instruction.comment.text), self.comment_width)
            else:
                rowspan = 1

class TableWriter:
    def __init__(self, max_width, min_col_width):
        self.max_width = max_width
        self.min_col_width = min_col_width
        self.cell_padding = len(' | ')
        self.table = None

    def format_table(self, text):
        table_parser = TableParser()
        self.table = table_parser.parse_table(text)
        self.table.cell_padding = self.cell_padding
        self.table.prepare_cells()
        self.table.reduce_width(self.max_width, self.min_col_width)
        return self._create_table_text()

    def _create_table_text(self):
        header_rows = [-1] + self.table.get_header_rows()
        lines = []
        prev_row = [(None, 1, 0)] * self.table.num_cols
        for row_index, row in enumerate(self.table.rows + [[]]):
            # Deal with remaining lines in previous row
            finished = row_index == 0
            while not finished:
                line = ''
                adj_transparent = True
                finished = True
                col_index = 0
                while col_index < self.table.num_cols:
                    prev_cell, prev_rowspan, prev_line_index = prev_row[col_index]
                    line += ' ' if adj_transparent and prev_cell.transparent else '|'
                    if prev_cell:
                        adj_transparent = prev_cell.transparent
                        colspan = prev_cell.colspan
                        contents = prev_cell.contents[prev_line_index] if prev_line_index < len(prev_cell.contents) else ''
                    else:
                        colspan = 1
                        contents = ''
                    if contents:
                        finished = False
                    line += " %s " % contents.ljust(self.table.get_cell_width(col_index, colspan))
                    prev_row[col_index] = (prev_cell, prev_rowspan, prev_line_index + 1)
                    col_index += colspan
                if not finished:
                    line += ' ' if prev_cell.transparent else '|'
                    lines.append(line)

            # Stop now if we've reached the end of the table (marked by an empty
            # row)
            if not row:
                break

            # Create a separator row if required
            if row_index - 1 in header_rows:
                lines.append(self._create_row_separator(row, row_index > 0))

            line = ''
            adj_transparent = True
            col_index = 0
            for cell in row:
                # Deal with previous rowspan > 1
                prev_cell, prev_rowspan, prev_line_index = prev_row[col_index]
                while prev_rowspan > 1:
                    line += ' ' if adj_transparent and prev_cell.transparent else '|'
                    adj_transparent, colspan = prev_cell.transparent, prev_cell.colspan
                    contents = prev_cell.contents[prev_line_index] if prev_line_index < len(prev_cell.contents) else ''
                    line += " %s " % contents.ljust(self.table.get_cell_width(col_index, colspan))
                    prev_row[col_index] = (prev_cell, prev_rowspan - 1, prev_line_index + 1)
                    col_index += colspan
                    prev_cell, prev_rowspan, prev_line_index = prev_row[col_index]

                # Deal with cells in this row
                line += ' ' if adj_transparent and cell.transparent else '|'
                adj_transparent, colspan = cell.transparent, cell.colspan
                line += " %s " % cell.contents[0].ljust(self.table.get_cell_width(col_index, colspan))
                prev_row[col_index] = (cell, cell.rowspan, 1)
                col_index += colspan

            # Deal with cells at the end of the previous row that have rowspan
            # greater than 1
            while col_index < len(prev_row):
                prev_cell, prev_rowspan, prev_line_index = prev_row[col_index]
                if prev_rowspan > 1:
                    line += ' ' if adj_transparent and prev_cell.transparent else '|'
                    adj_transparent, colspan = prev_cell.transparent, prev_cell.colspan
                    contents = prev_cell.contents[prev_line_index] if prev_line_index < len(prev_cell.contents) else ''
                    line += " %s " % contents.ljust(self.table.get_cell_width(col_index, colspan))
                    prev_row[col_index] = (prev_cell, prev_rowspan - 1, prev_line_index + 1)
                    col_index += prev_cell.colspan
                else:
                    break

            line += ' ' if cell.transparent and row_index == 0 else '|'
            lines.append(line)

        lines.append(self._create_row_separator(self.table.rows[-1]))
        return lines

    def _create_row_separator(self, row, fill=True):
        separator = ''
        col_index = 0
        adj_transparent = True
        for cell in row:
            solid = fill or not cell.transparent
            if col_index < cell.col_index:
                if col_index == 0:
                    separator = '+' if fill else ' '
                colspan = cell.col_index - col_index
                spacer = '-' if solid else ' '
                separator += spacer * (2 + self.table.get_cell_width(col_index, colspan))
            col_index, colspan = cell.col_index, cell.colspan
            separator += ' ' if adj_transparent and not solid else '+'
            spacer = '-' if solid else ' '
            separator += spacer * (2 + self.table.get_cell_width(col_index, colspan))
            adj_transparent = cell.transparent
            col_index += colspan
        return separator + ('+' if solid else '')
