#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2008-2011 Richard Dymond (rjdymond@gmail.com)
#
# This file is part of SkoolKit.
#
# SkoolKit is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# SkoolKit is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# SkoolKit. If not, see <http://www.gnu.org/licenses/>.

import sys
import time
from os.path import abspath, dirname, join
sys.path.insert(0, join(abspath(dirname(__file__)), 'lib'))
from skoolparser import SkoolParser, CASE_LOWER, CASE_UPPER, BASE_10, BASE_16
from skoolasm import SkoolWriter

DEF_INSTR_WIDTH=23

def clock(quiet, prefix, operation, *args, **kwargs):
    go = time.time()
    result = operation(*args, **kwargs)
    stop = time.time()
    if not quiet:
        sys.stderr.write('%s (%0.2fs)\n' % (prefix, stop - go))
    return result

def parse_options(options):
    quiet = verbose = debug = crlf = tab = False
    case = base = None
    instr_width = DEF_INSTR_WIDTH
    mode = 1
    fix_mode = 0
    do_warn = True
    html = False
    create_labels = False
    i = 0
    while i < len(options):
        option = options[i]
        if option == '-q':
            quiet = True
        elif option == '-v':
            verbose = True
        elif option == '-w':
            do_warn = False
        elif option == '-e':
            do_warn = False
            debug = True
        elif option == '-d':
            crlf = True
        elif option == '-l':
            case = CASE_LOWER
        elif option == '-u':
            case = CASE_UPPER
        elif option == '-D':
            base = BASE_10
        elif option == '-H':
            base = BASE_16
        elif option[:2] == '-i':
            instr_width = int(option[2:])
        elif option[:2] == '-f':
            fix_mode = int(option[2:])
        elif option == '-r':
            mode = 3
        elif option == '-s':
            mode = 2
        elif option == '-t':
            tab = True
        elif option == '-c':
            create_labels = True
        i += 1
    parser_mode = (case, base, mode, verbose, do_warn, fix_mode, html, create_labels)
    writer_mode = (crlf, tab, case == CASE_LOWER, instr_width)
    return parser_mode, writer_mode, quiet, debug

#
# Begin
#
if len(sys.argv) < 2:
    sys.stderr.write('Usage: %s [options] <skoolfile>\n\n' % sys.argv[0])
    sys.stderr.write('Options:\n')
    sys.stderr.write("  -q   Be quiet (don't show timings)\n")
    sys.stderr.write('  -v   Be verbose (show instruction substitutions)\n')
    sys.stderr.write('  -w   Suppress warnings\n')
    sys.stderr.write('  -e   Write processed lines to stderr (implies -w)\n')
    sys.stderr.write('  -d   Use CR+LF to end lines (default line terminator is LF)\n')
    sys.stderr.write('  -t   Use tab to indent instructions (default indentation is 2 spaces)\n')
    sys.stderr.write("  -l   Write disassembly in lower case\n")
    sys.stderr.write("  -u   Write disassembly in upper case\n")
    sys.stderr.write("  -D   Write disassembly in decimal\n")
    sys.stderr.write("  -H   Write disassembly in hexadecimal\n")
    sys.stderr.write('  -iN  Set instruction field width to N (default=%i)\n' % DEF_INSTR_WIDTH)
    sys.stderr.write('  -fN  Apply fixes:\n')
    sys.stderr.write('         N=0: None (default)\n')
    sys.stderr.write('         N=1: @ofix only\n')
    sys.stderr.write('         N=2: @ofix and @bfix\n')
    sys.stderr.write('         N=3: @ofix, @bfix and @rfix (implies -r)\n')
    sys.stderr.write('  -c   Create default labels for unlabelled instructions\n')
    sys.stderr.write('  -s   Use safe substitutions (@ssub)\n')
    sys.stderr.write('  -r   Use relocatability substitutions too (@rsub) (implies -f1)\n')
    sys.exit(1)

# Options
parser_mode, writer_mode, quiet, debug = parse_options(sys.argv[1:-1])

# Create the parser
skoolfile = sys.argv[-1]
parser = clock(quiet, 'Parsed %s' % skoolfile, SkoolParser, skoolfile, parser_mode, debug=debug)

# Write the ASM file
skool_writer = SkoolWriter(parser, *writer_mode)
clock(quiet, 'Wrote ASM to stdout', skool_writer.write)
