#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2010, 2011 Richard Dymond (rjdymond@gmail.com)
#
# This file is part of SkoolKit.
#
# SkoolKit is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# SkoolKit is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# SkoolKit. If not, see <http://www.gnu.org/licenses/>.

import sys
from os.path import abspath, dirname, join
sys.path.insert(0, join(abspath(dirname(__file__)), 'lib'))
from skoolparser import SkoolParser

BLOCKS = 'b'
BLOCK_TITLES = 't'
BLOCK_DESC = 'd'
REGISTERS = 'r'
BLOCK_COMMENTS = 'm'
SUBBLOCKS = 's'
COMMENTS = 'c'

class FakeInstruction:
    def __init__(self, address, comment):
        self.address = address
        self.comment = comment

def write_line(text):
    sys.stdout.write('%s\n' % text)

def get_instruction_ctl(op):
    if op.startswith('DEFB'):
        return 'b'
    if op.startswith('DEFW'):
        return 'w'
    if op.startswith('DEFM'):
        return 't'
    if op.startswith('DEFS'):
        return 'z'
    return 'c'

def _hex_address(address):
    return '$%04X' % int(address)

def _hex_number(num):
    integer = int(num)
    if integer < 10:
        fmt = '%i'
    elif integer < 256:
        fmt = '$%02X'
    else:
        fmt = '$%04X'
    return fmt % integer

def write_body(entry, elements, write_hex):
    # Split the entry by mid-routine comments
    entry_ctl = 'b' if entry.ctl in 'gu' else entry.ctl
    first_instruction = entry.instructions[0]
    if entry_ctl == 'i' and not first_instruction.operation:
        # Don't write any sub-blocks for an empty 'i' entry
        return
    blocks = [('', [first_instruction])]
    for instruction in entry.instructions[1:]:
        mrc = instruction.get_mid_routine_comment()
        if mrc:
            blocks.append((mrc, [instruction]))
        else:
            blocks[-1][1].append(instruction)

    for k, (mrc, instructions) in enumerate(blocks):
        if BLOCK_COMMENTS in elements and mrc:
            address = instructions[0].address
            write_line('D %s %s' % (_hex_address(address) if write_hex else address, mrc))

        if SUBBLOCKS not in elements:
            continue

        # Split the block by comment rowspans and instruction type
        sub_blocks = []
        i = 0
        prev_ctl = ''
        while i < len(instructions):
            instruction = instructions[i]
            comment = instruction.comment
            ctl = get_instruction_ctl(instruction.operation)
            if comment.rowspan > 1 or comment.text:
                inst_ctls = set()
                for inst in instructions[i:i + comment.rowspan]:
                    inst_ctls.add(get_instruction_ctl(inst.operation))
                if len(inst_ctls) > 1:
                    sub_blocks.append(('M', [FakeInstruction(instruction.address, instruction.comment)]))
                    for inst in instructions[i:i + comment.rowspan]:
                        inst.comment = None
                        sub_blocks.append((get_instruction_ctl(inst.operation), [inst]))
                else:
                    sub_blocks.append((ctl, instructions[i:i + comment.rowspan]))
                prev_ctl = ''
            elif sub_blocks and ctl == prev_ctl == 'c':
                sub_blocks[-1][1].append(instruction)
            else:
                sub_blocks.append((ctl, [instruction]))
                prev_ctl = ctl
            i += comment.rowspan

        # Write a ctl line for each sub-block
        for j, (ctl, instructions) in enumerate(sub_blocks):
            length = None
            offset = instructions[0].comment.rowspan + 1 if ctl == 'M' else 1
            if j + offset < len(sub_blocks):
                length = int(sub_blocks[j + offset][1][0].address) - int(instructions[0].address)
            elif k + 1 < len(blocks):
                length = int(blocks[k + 1][1][0].address) - int(instructions[0].address)
            comment = ''
            if instructions[0].comment and COMMENTS in elements:
                comment = instructions[0].comment.text
            if ctl == entry_ctl:
                if not comment:
                    # Don't write a ctl line for the first sub-block if it is
                    # commentless and its ctl is the same as the containing
                    # entry's
                    if j == 0:
                        continue
                    # Don't write a ctl line for a commentless sub-block of a
                    # code block
                    if ctl == 'c':
                        continue
                block_ctl = ' '
            else:
                block_ctl = ctl.upper()
            address = _hex_address(instructions[0].address) if write_hex else instructions[0].address
            lengths = ''
            if ctl == 'b' and len(instructions) > 1 and comment:
                # Find the lengths of each line in a multi-line 'B' sub-block
                # that has a comment
                defb_lengths = []
                for defb in instructions:
                    defb_lengths.append(defb.operation.count(',') + 1)
                length = sum(defb_lengths)
                while len(defb_lengths) > 1 and defb_lengths[-1] == defb_lengths[-2]:
                    defb_lengths.pop()
                lengths = ',' + ','.join([_hex_number(d) if write_hex else str(d) for d in defb_lengths])
            if length:
                if write_hex:
                    length = _hex_number(length)
                write_line(('%s %s,%s%s %s' % (block_ctl, address, length, lengths, comment)).rstrip())
            else:
                write_line(('%s %s %s' % (block_ctl, address, comment)).rstrip())

def write_entry(entry, elements, write_hex):
    address = _hex_address(entry.address) if write_hex else entry.address
    if BLOCKS in elements:
        if BLOCK_TITLES in elements:
            write_line(('%s %s %s' % (entry.ctl, address, entry.description)).rstrip())
        else:
            write_line('%s %s' % (entry.ctl, address))
    if BLOCK_DESC in elements:
        if entry.details:
            write_line('D %s %s' % (address, entry.details))
    if REGISTERS in elements:
        for reg in entry.registers:
            write_line('R %s %s %s' % (address, reg.name, reg.contents))
    write_body(entry, elements, write_hex)
    if BLOCK_COMMENTS in elements and entry.end_comment:
        write_line('E %s %s' % (address, entry.end_comment))

def write_ctls(entries, elements, write_hex):
    for entry in entries:
        write_entry(entry, elements, write_hex)

def parse_options(options):
    elements = BLOCKS + BLOCK_TITLES + BLOCK_DESC + REGISTERS + BLOCK_COMMENTS + SUBBLOCKS + COMMENTS
    write_hex = False
    i = 0
    while i < len(options):
        option = options[i]
        if option[:2] == '-w':
            elements = option[2:]
        elif option == '-h':
            write_hex = True
        i += 1
    return elements, write_hex

################################################################################
# Begin
################################################################################
if len(sys.argv) < 2:
    sys.stderr.write('Usage: %s [options] <skoolfile>\n' % sys.argv[0])
    sys.stderr.write('\nOptions:\n')
    sys.stderr.write("  -wX  Write only these elements, where X is one or more of:\n")
    sys.stderr.write("         %s = block types and addresses\n" % BLOCKS)
    sys.stderr.write("         %s = block titles\n" % BLOCK_TITLES)
    sys.stderr.write("         %s = block descriptions\n" % BLOCK_DESC)
    sys.stderr.write("         %s = registers\n" % REGISTERS)
    sys.stderr.write("         %s = mid-block comments and block end comments\n" % BLOCK_COMMENTS)
    sys.stderr.write("         %s = sub-block types and addresses\n" % SUBBLOCKS)
    sys.stderr.write("         %s = instruction-level comments\n" % COMMENTS)
    sys.stderr.write("  -h   Write addresses in hexadecimal format\n")
    sys.exit(1)

elements, write_hex = parse_options(sys.argv[1:-1])

# Create the parser
skoolfile = sys.argv[-1]
case = None
base = None
mode = 0
verbose = False
warn = False
fix_mode = 0
html = False
create_labels = False
mode_args = (case, base, mode, verbose, warn, fix_mode, html, create_labels)
parser = SkoolParser(skoolfile, mode_args)

# Write the ctl directives
write_ctls(parser.memory_map, elements, write_hex)
